/** 
 *   Refclock driver for Thunderbolt GPS Disciplined Clock
 *   
 *   (c) Copyright 2005 Fernando P. Hauscarriaga <fernandoph@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR) <www.iar.unlp.edu.ar>.
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined(SYS_WINNT)
#undef close
#define close closesocket
#endif

#if defined(REFCLOCK) && (defined(THUNDERBOLT) || defined(CLOCK_THUNDERBOLT))

#include "refclock_thunderbolt.h"

/* Table to get from month to day of the year */
const int days_of_year [12] = {
	0,  31,  59,  90, 120, 151, 181, 212, 243, 273, 304, 334
};

/*---------------------------------------------------------------------------*/
/*
 * thunderbolt_start - open the devices and initialize data for processing
 */

static int
thunderbolt_start (
#ifdef THUNDERBOLT
	unit, peer
	)
	int unit;
	struct peer *peer;
#else /* ANSI */
	int unit,
	struct peer *peer
	)
#endif
{
	struct thunderbolt_unit *up;
	struct refclockproc *pp;
	int fd;
	char gpsdev[20];

	struct termios tio;
#ifdef SYS_WINNT
	(void) sprintf(gpsdev, "COM%d:", unit);
#else	
	(void) sprintf(gpsdev, DEVICE, unit);
#endif
	/*
	 * Open serial port. 
	 */
#if defined THUNDERBOLT
	 fd = open(gpsdev, O_RDWR
#ifdef O_NONBLOCK
                  | O_NONBLOCK
#endif
                  );
#else /* NTP 4.x */
	fd = refclock_open(gpsdev, SPEED232, LDISC_RAW);
#endif
	if (fd <= 0) {
#ifdef DEBUG
                printf("Thunderbolt(%d) start: open %s failed\n", unit, gpsdev);
#endif
		       return 0;
		}

#if defined THUNDERBOLT 
        tio.c_cflag = (CS8|CLOCAL|CREAD);
        tio.c_iflag = (IGNBRK);
        tio.c_oflag = (0);
        tio.c_lflag = (0);

        if (cfsetispeed(&tio, SPEED232) == -1) { 
        msyslog(LOG_ERR,"Thunderbolt(%d) cfsetispeed(fd, &tio): %m",unit);
#ifdef DEBUG
                printf("Thunderbolt(%d) cfsetispeed(fd, &tio)\n",unit);
#endif
                return 0;
        }
        if (cfsetospeed(&tio, SPEED232) == -1) {
#ifdef DEBUG
                printf("Thunderbolt(%d) cfsetospeed(fd, &tio)\n",unit);
#endif
                msyslog(LOG_ERR,"Thunderbolt(%d) cfsetospeed(fd, &tio): %m",unit);
                return 0;
        }
#else /* NTP 4.x */
        if (tcgetattr(fd, &tio) < 0) {
                msyslog(LOG_ERR, 
			"Thunderbolt(%d) tcgetattr(fd, &tio): %m",unit);
#ifdef DEBUG
                printf("Thunderbolt(%d) tcgetattr(fd, &tio)\n",unit);
#endif
                return (0);
        }
	tio.c_iflag &= ~ICRNL;
#endif /*  NTP 4.x */

	if (tcsetattr(fd, TCSANOW, &tio) == -1) {
                msyslog(LOG_ERR, "Thunderbolt(%d) tcsetattr(fd, &tio): %m",unit);
#ifdef DEBUG
                printf("Thunderbolt(%d) tcsetattr(fd, &tio)\n",unit);
#endif
                return 0;
        }

	/*
	 * Allocate and initialize unit structure
	 */
	up = (struct thunderbolt_unit *) emalloc(sizeof(struct thunderbolt_unit));
	      
	if (!(up)) {
                msyslog(LOG_ERR, "Thunderbolt(%d) emalloc: %m",unit);
#ifdef DEBUG
                printf("Thunderbolt(%d) emalloc\n",unit);
#endif
		(void) close(fd);
		return (0);
	}

	memset((char *)up, 0, sizeof(struct thunderbolt_unit));


/* TODO:   Chequear todo esto de praecis!!  **********   */

	up->type = CLK_TYPE(peer);
	switch (up->type) {
		case CLK_TRIMBLE:
			/* Normal mode, do nothing */ 
			break;
		case CLK_PRAECIS:
			msyslog(LOG_NOTICE, "Palisade(%d) Praecis mode enabled\n",unit);
			break;
		default:
			msyslog(LOG_NOTICE, "Palisade(%d) mode unknown\n",unit);
			break;
	}
/*    TODO END   **************************      */

        pp = peer->procptr;
	pp->io.clock_recv = thunderbolt_io;
	pp->io.srcclock = (caddr_t)peer;
	pp->io.datalen = 0;
	pp->io.fd = fd;
	if (!io_addclock(&pp->io)) {
#ifdef DEBUG
                printf("Thunderbolt(%d) io_addclock\n",unit);
#endif
		(void) close(fd);
		free(up);
		return (0);
	}

	/*
	 * Initialize miscellaneous variables
	 */
	pp->unitptr = (caddr_t)up;
	pp->clockdesc = DESCRIPTION;

	peer->precision = PRECISION;
	peer->sstclktype = CTL_SST_TS_UHF;
	peer->minpoll = TRMB_MINPOLL;
	peer->maxpoll = TRMB_MAXPOLL;
	memcpy((char *)&pp->refid, REFID, 4);
	
	up->leap_status = 0;
	up->unit = (short) unit;
	up->rpt_status = TSIP_PARSED_EMPTY;
    	up->rpt_cnt = 0;

	return 1;
}

/*---------------------------------------------------------------------------*/
/*
 * thunderbolt_shutdown - shut down the clock
 */

static void
thunderbolt_shutdown (
#ifdef PALISADE
	unit, peer
	)
	int unit;
	struct peer *peer;
#else /* ANSI */
	int unit,
	struct peer *peer
	)
#endif
{
	struct thunderbolt_unit *up;
	struct refclockproc *pp;
	pp = peer->procptr;
	up = (struct thunderbolt_unit *)pp->unitptr;
	io_closeclock(&pp->io);
	free(up);
}


/*---------------------------------------------------------------------------*/
/* 
 * unpack_date - get day and year from date
 */

int
day_of_year (
#ifdef THUNDERBOLT
	dt
	)
	char * dt;
#else
	char * dt
	)
#endif
{
	int day, mon, year;

	mon = dt[1];

       /* Check month is inside array bounds */
       if ((mon < 1) || (mon > 12))
		return -1;

	day = dt[0] + days_of_year[mon - 1];
	year = getint((u_char *) (dt + 2)); 
	if ( !(year % 4) && ((year % 100) || 
		(!(year % 100) && !(year%400)))
			&&(mon > 2))
			day ++; /* leap year and March or later */

	return day;
}

/*---------------------------------------------------------------------------*/
/* 
 * TSIP_decode - decode the TSIP data packets 
 */
int
TSIP_decode (
#ifdef THUNDERBOLT
	peer
	)
	struct peer *peer;
#else
	struct peer *peer
	)
#endif
{
        int st;
	long   secint, tow;
	double secs;
	double secfrac;
	unsigned short event = 0;
	
	struct thunderbolt_unit *up;
	struct refclockproc *pp;

	pp = peer->procptr;
	up = (struct thunderbolt_unit *)pp->unitptr;

	/*
	 * Check the time packet, decode its contents. 
	 * If the timecode has invalid length or is not in
	 * proper format, declare bad format and exit.
	 */
	
	/*
	 * We cast both to u_char to as 0x8f uses the sign bit on a char
	 */
	if ((u_char) up->rpt_buf[0] == (u_char) 0x8f) {
	/* 
	 * Superpackets
	 */
	   event = (unsigned short) (getint((u_char *) &mb(1)) & 0xffff);
	   if (!((pp->sloppyclockflag & CLK_FLAG2) || event)) 
		/* Ignore Packet */
			return 0;	   
	
	   switch (mb(0) & 0xff) {
	     int GPS_UTC_Offset;
             float osci_offset;
	     double n_sec;
             
             case PACKET_8FAC:   
	     if (up->polled <= 0)
                        return 0; 

                if (up->rpt_cnt != LENCODE_8FAC)/* check length */
                        break;

#ifdef DEBUG
if (debug > 1) {
                double lat, lon, alt;
                lat = getdbl((u_char *) &mb(36)) * R2D;
                lon = getdbl((u_char *) &mb(44)) * R2D;
                alt = getdbl((u_char *) &mb(52));

                printf("TSIP_decode: unit %d: Latitude: %03.4f Longitude: %03.4f Alt: %05.2f m\n",
                                up->unit, lat,lon,alt);
                printf("TSIP_decode: unit %d\n", up->unit);
               }
#endif
		if (getint((u_char *) &mb(10)) & 0x80) {
		   pp->leap = LEAP_ADDSECOND;  /* we ASSUME addsecond */
                   printf("Leap ADDSECOND...\n");
		}
		else {
		   pp->leap = LEAP_NOWARNING;
                   printf("Leap NOWARNING...\n");
		}

		if (debug > 1) {
		   printf("TSIP_decode: unit %d: 0x%02x leap %d\n",
			  up->unit, mb(0) & 0xff, pp->leap);
		}
		if (debug > 1) {
printf("** %cReceiver MODE: 0x%02X\n", 7, (u_char)mb(1));
if (mb(1) == 0x00)
     printf("                AUTOMATIC\n");
if (mb(1) == 0x01)
     printf("                SINGLE SATELLITE\n");
if (mb(1) == 0x03)
     printf("                HORIZONTAL(2D)\n");
if (mb(1) == 0x04)
     printf("                FULL POSITION(3D)\n");
if (mb(1) == 0x05)
     printf("                DGPR REFERENCE\n");
if (mb(1) == 0x06)
     printf("                CLOCK HOLD(2D)\n");
if (mb(1) == 0x07)
     printf("                OVERDETERMINED CLOCK\n");

printf("\n** Disciplining MODE 0x%02X:\n", (u_char)mb(2));
if (mb(2) == 0x00)
     printf("                NORMAL\n");
if (mb(2) == 0x01)
     printf("                POWER-UP\n");
if (mb(2) == 0x02)
     printf("                AUTO HOLDOVER\n");
if (mb(2) == 0x03)
     printf("                MANUAL HOLDOVER\n");
if (mb(2) == 0x04)
     printf("                RECOVERY\n");
if (mb(2) == 0x6)
     printf("                DISCIPLINING DISABLED\n");
		}
		   
                return 0;
                break;

          case PACKET_8FAB:
                /* Thunderbolt Primary Timing Packet */

                if (up->rpt_cnt != LENCODE_8FAB) /* check length */
                        break;

                if (up->polled  <= 0)
                        return 0;

		GPS_UTC_Offset = getint((u_char *) &mb(7));
                
                if (GPS_UTC_Offset == 0){ /* Check UTC Offset */
#ifdef DEBUG
		      printf("TSIP_decode: UTC Offset Unknown\n");
#endif
			  break;
		 }

		pp->nsec = 0;
#ifdef DEBUG		
		printf("\nTiming Flags are:\n");
		printf("timing_flag is: 0x%X\n", mb(9));
		if ((mb(9) & 0x01) != 0)
			printf ("	Getting UTC time\n");
		else
			printf ("	Getting GPS time\n");
		if ((mb(9) & 0x02) != 0)
			printf ("	PPS is from UTC\n");
		else
			printf ("	PPS is from GPS\n");
		if ((mb(9) & 0x04) != 0)
			printf ("	Time is not Set\n");
		else
			printf ("	Time is Set\n");
		if ((mb(9) & 0x08) != 0)
			printf("	I dont have UTC info\n");
		else
			printf ("	I have UTC info\n");
		if ((mb(9) & 0x10) != 0)
			printf ("	Time is from USER\n\n");
		else
			printf ("	Time is from GPS\n\n");	
#endif		

                if ((pp->day = day_of_year(&mb(13))) < 0)
                  break;
                
		tow = getlong((u_char *) &mb(1));
		printf("********** TOW: %ld\n", tow);
                printf("****** DAY: %d\n", pp->day);
                pp->year = getint((u_char *) &mb(15));
                pp->hour = mb(12);
                pp->minute = mb(11);
                pp->second = mb(10);

#ifdef DEBUG
        if (debug > 1)
printf("** TSIP_decode: unit %d: %02X #%d %02d:%02d:%02d.%06ld %02d/%02d/%04d UTC: %02d \n",up->unit, mb(0) & 0xff, event, pp->hour, pp->minute, pp->second, pp->nsec, mb(14), mb(13), pp->year, GPS_UTC_Offset);
#endif
                return 1;
                break;

             case PACKET_41:
#ifdef DEBUG
	        osci_offset = getfloat((u_char *) &mb(11));
                printf("** Oscillator Offset: %f\n", osci_offset);
#endif
             return 0;
             break;
        
	  default: 	
		/* Ignore Packet */
		return 0;}
/* switch */
	}/* if 8F packets */

else if (up->rpt_buf[0] == (u_char)0x42)
{
  printf("0x42\n");
  return 0;
}
else if (up->rpt_buf[0] == (u_char)0x43)
{
  printf("0x43\n");
  return 0;
}
else if (up->rpt_buf[0] == PACKET_41)
{
  printf("0x41\n");
  return 0;
}
else if (up->rpt_buf[0] == PACKET_6D){
#ifdef DEBUG
		int sats;

if ((mb(0) & 0x01) && (mb(0) & 0x02))
     printf("2d Fix Dimension\n");
if (mb(0) & 0x04)
     printf("3d Fix Dimension\n");

		if (mb(0) & 0x08)
			printf("Fix Mode is MANUAL\n");
		else
			printf("Fix Mode is AUTO\n");
		
                sats = mb(0) & 0xF0;
		sats = sats >> 4;
		printf("Tracking %d Satellites\n", sats);
#endif
		return 0;
        } /* else if not super packet */
refclock_report(peer, CEVNT_BADREPLY);
up->polled = -1;
#ifdef DEBUG

printf("TSIP_decode: unit %d: bad packet %02x-%02x event %d len %d\n", 
  		  up->unit, up->rpt_buf[0] & 0xff, mb(0) & 0xff, 
	  	  event, up->rpt_cnt);
#endif
return 0;

/*---------------------------------------------------------------------------*/
/*
 * thunderbolt__receive - receive data from the serial interface
 */

static void
thunderbolt_receive (
#ifdef THUNDERBOLT
	peer
	)
	struct peer * peer;
#else /* ANSI */
	struct peer * peer
	)
#endif
{
	struct thunderbolt_unit *up;
	struct refclockproc *pp;

	/*
	 * Initialize pointers and read the timecode and timestamp.
	 */
	pp = peer->procptr;
	up = (struct thunderbolt_unit *)pp->unitptr;
		
	if (! TSIP_decode(peer)) return;
	
	if (up->polled <= 0) 
	    return;   /* no poll pending, already received or timeout */

	up->polled = 0;  /* Poll reply received */
	pp->lencode = 0; /* clear time code */
#ifdef DEBUG
	if (debug) 
		printf(
	"thunderbolt_receive: unit %d: %4d %03d %02d:%02d:%02d.%06ld\n",
			up->unit, pp->year, pp->day, pp->hour, pp->minute, 
			pp->second, pp->nsec);
#endif

	/*
	 * Process the sample
	 * Generate timecode: YYYY DoY HH:MM:SS.microsec 
	 * report and process 
	 */

	(void) sprintf(pp->a_lastcode,"%4d %03d %02d:%02d:%02d.%06ld",
		   pp->year,pp->day,pp->hour,pp->minute, pp->second,pp->nsec); 
	pp->lencode = 24;

#ifdef THUNDERBOLT
    	pp->lasttime = current_time;
#endif
	if (!refclock_process(pp
#ifdef THUNDERBOLT
		, THUNDERBOLT_SAMPLES, THUNDERBOLT_SAMPLES * 3 / 5
#endif
		)) {
		refclock_report(peer, CEVNT_BADTIME);

#ifdef DEBUG
		printf("thunderbolt_receive: unit %d: refclock_process failed!\n",
			up->unit);
#endif
		return;
	}

	record_clock_stats(&peer->srcadr, pp->a_lastcode); 

#ifdef DEBUG
        if (debug)
	    printf("thunderbolt_receive: unit %d: %s\n",
		   up->unit, gmprettydate(&pp->lastrec));
#endif
	pp->lastref = pp->lastrec;
	refclock_receive(peer
#ifdef PALISADE
		, &pp->offset, 0, pp->dispersion,
              &pp->lastrec, &pp->lastrec, pp->leap		
#endif		
		);
}

/*---------------------------------------------------------------------------*/
/*
 * thunderbolt_poll - called by the transmit procedure
 *
 */
static void
thunderbolt_poll (
#ifdef THUNDERBOLT
	unit, peer
	)
	int unit;
	struct peer *peer;
#else
	int unit,
	struct peer *peer
	)
#endif
{
	struct thunderbolt_unit *up;
	struct refclockproc *pp;
	
	pp = peer->procptr;
	up = (struct thunderbolt_unit *)pp->unitptr;

	pp->polls++;
	if (up->polled > 0) /* last reply never arrived or error */ 
	    refclock_report(peer, CEVNT_TIMEOUT);

	up->polled = 2; /* synchronous packet + 1 event */
	
#ifdef DEBUG
	if (debug)
	    printf("thunderbolt_poll: unit %d: polling %s\n", unit,
		   (pp->sloppyclockflag & CLK_FLAG2) ? 
			"synchronous packet" : "event");
#endif 

	if (pp->sloppyclockflag & CLK_FLAG2) 
	    return;  /* using synchronous packet input */

	if(up->type == CLK_PRAECIS) {
		if(write(peer->procptr->io.fd,"SPSTAT\r\n",8) < 0)
			msyslog(LOG_ERR, "Thunderbolt(%d) write: %m:",unit);
		else {
			praecis_msg = 1;
			return;
		}
	}

	if (HW_poll(pp) < 0) 
	    refclock_report(peer, CEVNT_FAULT); 
}

/*---------------------------------------------------------------------------*/

/*------------------------------------------------------------
 *
 *  Que sera esto, no?? ...
 */
static void
praecis_parse(struct recvbuf *rbufp, struct peer *peer)
{
	static char buf[100];
	static int p = 0;
	struct refclockproc *pp;

	pp = peer->procptr;
	

	memcpy(buf+p,rbufp->recv_space.X_recv_buffer, rbufp->recv_length);
	p += rbufp->recv_length;

	if(buf[p-2] == '\r' && buf[p-1] == '\n') {
		buf[p-2] = '\0';
		record_clock_stats(&peer->srcadr, buf);

		p = 0;
		praecis_msg = 0;

		if (HW_poll(pp) < 0)
			refclock_report(peer, CEVNT_FAULT);

	}
}
/*---------------------------------------------------------------*/




